#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <ctype.h>
#include <pwd.h>

#ifdef X11
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#else
#include <vga.h>
#endif

#include "tetX.h"
#include "egafont.h"
#include "keyjoy.h"
#ifdef X11
XImage *DrawCharOnXImage(tchar textChar, unsigned long fgrPix, unsigned long bgrPix);
#endif

#ifndef SIGCLD
#define SIGCLD SIGCHLD
#endif

extern int  ScreenWidth, ScreenHeight, ScreenDepth, ScrollAvailable;
extern int  BoardX, BoardY, BoardW, BoardH;
extern int  NextX, NextY, NextW, NextH;
extern int  JoyX, JoyY, JoyW, ShadowY;
extern int  BkgrColour, TxtColour;
extern int  Level, Demo;
extern int  ScreenBusy, TimerSuspended;
extern long Ticks;
extern char *PlayListFileName;
extern char FillPats[][8];
extern WHERE_WE_ARE where;
extern BOTTOM_MODE  BottomMode;
extern int  VideoChipset;
int    ScrollColour, ScrollerBusy;


int ScrollX, ScrollY, ScrollW, ScrollH;
int ScrollVScale, ScrollHScale;
int ScrollEX, ScrollEY, ScrollEW, ScrollEH;


static int BufferH, BufferW;

static char *TextPtr=NULL;
static int Suspend;
static char *ScrollBuffer = NULL;
#ifndef SCROLLER_DISABLED
static int gran = 2;
static short kill_player;
static short redir_input, redir_output, redir_error;


#ifdef X11
extern	Display	*display;
extern	Visual	*visual;
extern	Window	window;
extern  Pixmap  screenContext;
extern  GC	graphContext;
extern  unsigned long BkgrXColour;
extern  unsigned long PixValues[16];

static unsigned long ScrollXColour;
static XImage *ScrollerImage = NULL;
static XImage *ScrollerLetterImage = NULL;
static int  ScrollLetPos, ScrollLetWid;
#else
static int  ScrollCurColour;
static int  ScrollMask;
static char *FontPtr;
#endif
static int ScrollSpeed, SpeedCount;
static int ScrollTickBusy;
static int  randTab[20];
static int  randTabPos;
static char *CurPtr=NULL, **RandTextPtr;
static int  CleanCount, RandCap;
#endif


/* ************************************************** */
/*       Sleep - the most problematic routine         */
/* ************************************************** */

void sleep_until(long StopAt)
{
   while ((!TimerSuspended) && (Ticks < StopAt))
   {  pause();
#ifdef X11
	  kbdGet();
#else
      while(vga_getkey() != 0);	/* Clear keyboard buffer */
#endif
   }
}    

void RefreshAndWait(long interv, RECT *rc)
{ 
    ContextRectToScreen(rc);
    sleep_until(Ticks+interv);    
}    
   

/* ************************************************** */
/*                       SCROLLER                     */
/* ************************************************** */

#if defined (X11) && !defined(SCROLLER_DISABLED)
static int CreateScrollerImage()
{
    char *data;
    unsigned size;

    ScrollerImage = XCreateImage(display, visual, ScreenDepth, ZPixmap,
                                  0, NULL, BufferW, BufferH*ScrollVScale,
				  8, 0);

    size = ScrollerImage->bytes_per_line*ScrollerImage->height;
    data = malloc(size);
    if (!data) goto BadLuck;

    ScrollerImage->data = data;

    return 0;

BadLuck:
    if (ScrollerImage) XDestroyImage(ScrollerImage);
    return 4;
}
#endif

int InitScroller(void)
{
   ScrollHScale = 1;
   ScrollVScale = (ScreenHeight > 350) ? 2 : 1;

   ScrollW = ScreenWidth - (ScrollX << 1);

   BufferH = egaGetFontHeight();

   ScrollEH = BufferH*ScrollVScale;

   ScrollH = ScrollEH;
   if (ScreenHeight > 350) ScrollH += 8;

   ShadowY = BoardY + BoardH + 8;

   if (ScreenHeight <= 350) ScrollY = ShadowY;
   else
   {  if (ScreenHeight < 480) ShadowY += 7;           /* X11 case */
      ScrollY = (ScreenHeight+BoardY+BoardH-ScrollH) >> 1;
   }

   ScrollEY = ScrollY + ((ScrollH-ScrollEH) >> 1);

   ScrollColour = SCROLL_COLOUR;
   TextPtr = NULL;

   return 0;
}

int SetScrollerParam(int NewWidth)
{

    if (ScrollBuffer) free(ScrollBuffer);
    ScrollW = NewWidth;

    ScrollX = (ScreenWidth - ScrollW) >> 1;

    BufferW = ScrollEW = ScrollW - 16;

    ScrollEX = ScrollX + 8;

#if defined(SCROLLER_DISABLED)
    ScrollBuffer = NULL;
    return  4;
#elif  defined(X11)
    ScrollBuffer = NULL;
    return CreateScrollerImage();
#else
    ScrollBuffer = (char *) calloc(BufferH, BufferW);
    if (!ScrollBuffer) return 4;
    return 0;
#endif
}


int DeinitScroller(void)
{
#if defined(SCROLLER_DISABLED)
    return 0;
#else
#if defined(X11)
   if (ScrollerImage) XDestroyImage(ScrollerImage);
   if (ScrollerLetterImage)
   {  XDestroyImage(ScrollerLetterImage);
#else
   if (ScrollBuffer)
   {  free(ScrollBuffer);
#endif
      return 0;
   }
   else
      return 4;
#endif
}

#ifndef SCROLLER_DISABLED
static void AdvanceScroll(void);
#endif

void ShowScrollerFrame(void)
{
    RECT rc;

    rc.top = ScrollY; rc.bottom = ScrollY + ScrollH;
    rc.left = ScrollX; rc.right = ScrollX + ScrollW;


    if (where == W_PROLOGUE || ((BottomMode == B_SCROLLER) && Demo) || BottomMode == B_HELP)
           DrawFramedRect(&rc);
    else
    if (where == W_GAME)
    {  rc.top  -= 6;
       rc.bottom += 4;
       if (ScreenHeight > 350) {rc.top -=5; rc.bottom ++; }
       rc.left -= 8;  rc.right += 8;
       DrawPatRect(&rc, FillPats[Level], 5);


       if (ScreenHeight <= 350)
       {  rc.top = BoardY; rc.left = BoardX;
          rc.bottom = BoardY + BoardH;
          rc.right = BoardX + BoardW;
          EncompassFramedRect(&rc);

          rc.left = JoyX; rc.right = JoyX + JoyW;
          rc.top = JoyY;
          EncompassFramedRect(&rc);
       }

    }

}

#ifndef SCROLLER_DISABLED
static void RandomizeTable(int size)
{  int i, j, k;
   int *ptr1, *ptr2;

   for (i=0; i<size; i++) randTab[i] = i;

   for (i=0; i<size-1; i++)
   {  k = size-i;
      j = get_rand(k--);
      /* I believe, it's faster than access by index */
      if (j != k)
      { ptr1 = randTab+k;
        ptr2 = randTab+j;
        *ptr1 ^= *ptr2;
        *ptr2 ^= *ptr1;
        *ptr1 ^= *ptr2;
      }
   }
}
#endif

int StartScroller(char *text, int speed, int randomN,  char *randomText[] )
{
#ifdef SCROLLER_DISABLED
    return 4;
#else

#ifdef X11
    int x, y;

//    if (!ScrollAvailable || !ScrollerImage)  return 4;
    if (!ScrollerImage)  return 4;
    ScrollXColour = PixValues[ScrollColour];
    for (x=0; x<BufferW; x++)
        for (y=0; y<BufferH * ScrollVScale; y++)
            XPutPixel(ScrollerImage, x, y, BkgrXColour);

#else
//    if (!ScrollAvailable || !ScrolBuffer)  return 4;
    if (!ScrollBuffer) return 4;
    bzero(ScrollBuffer, BufferH*BufferW);
    ScrollCurColour = ScrollColour;
    ScrollMask = 0;
#endif

    Suspend = 1;
    ShowScrollerFrame();

    SpeedCount = ScrollSpeed = speed;

    RandCap = randomN;
    if (randomN > 0) RandomizeTable(randomN);
    randTabPos = 0;
    RandTextPtr = randomText;
    Suspend = 0;
    ScrollerBusy = 0;
    ScrollTickBusy = 0;

    CurPtr = TextPtr = text;

    return 0;
#endif
}

#ifndef SCROLLER_DISABLED
static void AdvanceScroll(void)
{   int i, k, j, y,  inc;

#ifdef X11
    unsigned long Colour;
#else
    char fchar;
    int  mask;
    char *fptr, *bptr;
#endif

    if (CurPtr==NULL)   // Cleaning
    {  if (CleanCount <= 0)
       { if (RandCap == 0)
         {  TextPtr = NULL;
            return;
         }
         else
         { // TextPtr = RandTextPtr[get_rand(RandCap)];
            TextPtr = RandTextPtr[randTab[randTabPos]];
            randTabPos = (randTabPos + 1) % RandCap;
            CurPtr = TextPtr;
#ifdef X11
            ScrollXColour = PixValues[ScrollColour];
#else
            ScrollCurColour = ScrollColour;
            ScrollMask = 0;
#endif
         }
      /*
         else
         {  CurPtr = TextPtr;
            ScrollMask = 0;
            SpeedCount = ScrollDelay;
         }
      */
       }
       else
         CleanCount -= gran;
    }

#ifdef X11
    if(CurPtr && ScrollLetPos>=ScrollLetWid)
#else
    if(CurPtr && ScrollMask==0)
#endif
    {  unsigned char c;

#ifdef X11
       if (ScrollerLetterImage) XDestroyImage(ScrollerLetterImage);
       ScrollerLetterImage = NULL;
#endif
       c = *CurPtr++;
       if (c == '\x7F')
       {
#ifdef X11
           ScrollXColour = PixValues[(int)(*CurPtr++)];
#else
           ScrollCurColour = *CurPtr++;
#endif
           c = *CurPtr++;
       }

       if (c==0)
       {   CurPtr = NULL;
           CleanCount = BufferW;
       }
       else
       {
#ifdef X11
    	 ScrollerLetterImage = DrawCharOnXImage(c, ScrollXColour, BkgrXColour);
	     ScrollLetWid = ScrollerLetterImage->width;
  	     ScrollLetPos = 0;
#else
		 FontPtr = egaGetFontPtr(c);
	     ScrollMask = 0x80;
#endif
       }
    }

    inc = BufferW-gran;

    ScrollerBusy = 1;


#ifdef X11
     for (i=0; i < BufferH; i++)
     {

	   y = i*ScrollVScale;

	   /* Get prev line */
           for (k=0; k<inc; k++)
	       XPutPixel(ScrollerImage, k, y,
	                  XGetPixel(ScrollerImage, k+gran, y));

	   k = ScrollLetPos;

	   Colour = BkgrXColour;
	   for  (j= 0; j<gran; j++)
	   {   if (CurPtr) Colour = XGetPixel(ScrollerLetterImage, k++, i);
	       XPutPixel(ScrollerImage, inc+j, y, Colour);
	   }

	   /* Copy down */
	   j = ScrollVScale;
	   while(--j >= 1)
	      for (k=0; k<inc; k++)
	         XPutPixel(ScrollerImage, k, y+j,
		            XGetPixel(ScrollerImage, k, y));

      }

      if (!ScreenBusy)
          XPutImage(display, window, graphContext, ScrollerImage,
    	          0, 0, ScrollEX, ScrollEY, BufferW, BufferH*ScrollVScale);

      ScrollLetPos += gran;

#else

    // Adjust and display scroll buffer

      fptr = FontPtr;
      y = ScrollEY;
      bptr = ScrollBuffer;
      for (i=0; i<BufferH; i++)
      {
         memcpy (bptr, bptr+gran, inc);
         if (CurPtr) fchar = *fptr++; else fchar = 0;
         mask = ScrollMask;
         for (k=0; k<gran; k++)
         {  *(bptr+inc+k) = (fchar & mask) ? ScrollCurColour : BkgrColour;
            mask >>= 1;
         }
         for (j=0; j<ScrollVScale; j++)
              vga_drawscansegment(bptr, ScrollEX, y++, ScrollEW);
//           setrow(ScrollEX, y++, ScrollEW, bptr);
         bptr += (inc+gran);
     }

     ScrollMask >>= gran;
#endif

    ScrollerBusy = 0;

    SpeedCount = ScrollSpeed;
}
#endif

void ScrollerTick(void)
{

#ifdef SCROLLER_DISABLED
    return;
#else
    if (TextPtr && !Suspend && !ScreenBusy && !ScrollTickBusy && --SpeedCount <= 0)
    {  ScrollTickBusy = 1;
       AdvanceScroll();
       ScrollTickBusy = 0;
    }
    return;
#endif

}

int IsScrollerActive()
{  return (TextPtr != NULL); }

void SuspendScroller(void)
{

#ifndef SCROLLER_DISABLED
   if (!Suspend)
   { while(ScrollTickBusy && !TimerSuspended) pause();
     Suspend = 1;
   }
#else
   ;
#endif
}

void ResumeScroller(void)
{
    Suspend = 0;
}

void StopScroller(void)
{
#ifndef SCROLLER_DISABLED
    if (!TextPtr) return;
    while(ScrollTickBusy && !TimerSuspended)
                         pause();
#ifdef X11
    if (ScrollerLetterImage)
    {  XDestroyImage(ScrollerLetterImage);
       ScrollerLetterImage = NULL;
    }
#endif
    TextPtr = NULL;
    ShowScrollerFrame();
#endif
    ;
}

/* ************************************************** */
/*                 BACKGROUND PLAYER                  */
/* ************************************************** */
#define mus_delay(interval) sleep_until(Ticks+(interval))

extern int  LoopMode, KillPlayMode; 
extern int  RedirInpMode, RedirOutMode, RedirErrMode;
static int  LoopNeeded;
#define MAX_PLAYLINES 20
static int  NumPlayLines;
static int  ChildPid = 0;
static char *PlayLines[MAX_PLAYLINES];
static char line_being_played[121];
static char *player_arg[41];


static void PlayBGEndHandler(int code);


static int StartForkedPlayer(int synch)
{  int chpid;

   signal(SIGCLD, PlayBGEndHandler);
   chpid = vfork();
   if (chpid)
   {  if (chpid == -1) chpid = 0;
      else 
      if(synch) mus_delay(LOAD_DELAY); 
      return (ChildPid = chpid);
   }

   signal(SIGALRM, SIG_IGN);

   if (redir_output)
       freopen("/dev/null", "w+", stdout); 

   if (redir_error)
       freopen("/dev/null", "w+", stderr); 

   if (redir_input)
       freopen("/dev/null", "r", stdin); 

   nice(4); 
   return execvp(player_arg[0], player_arg); 
}


static void PlayBGEndHandler(int code)
{
   
   if (ChildPid)
   {  int status; 
      waitpid(ChildPid, &status, 0); 
	  
	  if (WIFSIGNALED(status)) LoopNeeded = 0;
	  ChildPid = 0;
   }	   
 		
   if (LoopNeeded)  StartForkedPlayer(0);
}

void  process_modes(char *modes)
{     char c;
      int flag;

   while ((c=*modes++) != '\0')
   { 
        flag = isupper(c);

        switch (toupper(c))
        {  case  'K':
                KillPlayMode = flag; break;     
   
            case 'I':
	        RedirInpMode = flag; break;

            case 'O':
	        RedirOutMode = flag; break;

            case 'E':
	        RedirErrMode = flag; break;

            case 'L':
	        LoopMode = flag; break;
        }
    }

}

static int PlayFileInBackground(char *comlin)
{  int  i, flag;
   char *ptr;
   char c;    
   

   strncpy(line_being_played, comlin, 121);
   line_being_played[120] = '\0';
   ptr = line_being_played;

   kill_player = KillPlayMode; 
   redir_input = RedirInpMode;
   redir_output = RedirOutMode;
   redir_error = RedirErrMode;

   LoopNeeded = LoopMode;

   c = *ptr;
   if (c == '!')
   {  
      ptr++;
      kill_player = 1;
   }
   else
   if (c == '<')
   {  
      while ((c=*++ptr) != '>')
      { 
	 if (c == '\0') return -1;	

         flag = isupper(c);
         switch (toupper(c))
         {  case  'K':
               kill_player = flag; break;     
   
            case 'I':
	       redir_input = flag; break;

            case 'O':
	       redir_output = flag; break;

            case 'E':
	       redir_error = flag; break;

            case 'L':
	       LoopNeeded = flag; break;
         }
      }
	
      ptr++;  /* bypasing '>' */
   }      
      	      
   ptr += strspn(ptr, " ");

   ptr = strtok(ptr, " \t");
   if (ptr == NULL) return -1;


   player_arg[0] = ptr;

   for (i=1; i<19; )
   {  ptr = strtok(NULL, " \t");
      if (!ptr) break;
      if (!*ptr) continue;
      player_arg[i++] = ptr;
   }

   if (i >= 19) return -1;

   player_arg[i] = NULL;

   i = StartForkedPlayer(1);
   return i;

}


int StopBGPlay(int wait_needed)
{

   LoopNeeded = 0;
  
   if (ChildPid && kill(ChildPid, 0) == 0)
   {	long NextTicks;	

       if (kill_player)	 
          kill(ChildPid, SIGKILL);
       else		  
    	  kill(ChildPid, SIGTERM);

       if (wait_needed)
          mus_delay(KILL_DELAY);

	NextTicks = Ticks + 20;
	while (ChildPid  && Ticks < NextTicks)
		      pause();	   
	   //      waitpid(ChildPid, NULL, WUNTRACED);
	   
	if (kill(ChildPid, 0) == 0) return 4;
   }

   return 0;
}

int PlayPlayLine(int index, int mode)
{
    int ii;
    char *lineptr;

   if(index < 0 ||
      index >= NumPlayLines) return 0;

   if ( (mode & MODE_CHECK) && ChildPid) return 0;

   lineptr = PlayLines[index];
   if (*lineptr == '*')
   {
     if ((mode & MODE_FORCE) == 0) return 0;
     while(--index >= 0)
     {  lineptr = PlayLines[index];
        if (*lineptr != '*') break;
     }

     if (index < 0) return 0;
     
   }

   StopBGPlay(1);
   if (*lineptr == '-') return 0;

   ii = PlayFileInBackground(lineptr);
   return ii;

}

char *GetFullFileName(const char *filename)
{  char *fname;
   char *home = NULL;
   int  l;

   if (!filename) return NULL;

   l = 0;

   if (*filename == '~')
   {  int ul;

      filename++;


      fname = strchr(filename, '/');
      if (!fname) goto Adelante;

      ul = fname - filename;

      if (!ul)
          home = getenv("HOME");   // User home directory
      else
      {   struct passwd *pwent;

          home = NULL;

          fname = malloc(ul+1);
          if (!fname) goto Adelante;

          memcpy(fname, filename, ul);
          fname[ul] = '\0';
          pwent = getpwnam(fname);
          free(fname);

          if (pwent)
          {  home = pwent->pw_dir;
             if (home) filename += ul;
          }

      }

      if (home) l=strlen(home);
   }

Adelante:
   l += strlen(filename) + 1;

   fname = (char *) malloc(l);

   if (fname)
   { if (home)
     { strcpy(fname, home);
       strcat(fname, filename);
     }
     else
       strcpy(fname, filename);
   }

   return fname;
}

int InitPlayLines(void)
{
    FILE *f;
    int  rc = 8;
    char line[81];
    char *lineptr;

    player_arg[0] = NULL;
    
    if (!PlayListFileName) return 8;
    NumPlayLines = 0;

    f = fopen(PlayListFileName, "r");

    if (!f)
    {  printf ("\nERROR opening play list file %s\n", PlayListFileName);
       printf ("File may be missing.\n");
       free(PlayListFileName);
       PlayListFileName = NULL;	
       goto Exit;
    }

    while (NumPlayLines < MAX_PLAYLINES && !feof(f) && fgets(line, 80, f))
    {  // Getting rid of comments and endline chars
       // strtok (line, "#\n\r"); Doesn't work for some strange reasons
       lineptr = strchr(line, '#');
       if (!lineptr) lineptr = strchr(line, '\r');   // Just in case
       if (!lineptr) lineptr = strchr(line, '\n');
       if (lineptr) *lineptr = '\0';

       // Ignoring empty lines
       if (strcspn(line, " \t") == 0) continue;

       lineptr = strdup(line);
       if (!lineptr)
       { printf("\nNot enough memory to hold the play list\n.");
         goto Exit;
       }

       PlayLines[NumPlayLines++] = lineptr;
    }

    rc = 0;

Exit:
    if (f) fclose(f);
    if (rc)
    {  printf ("No background music will be played.\n\n");
       PlayListFileName = NULL;
    }

    return rc;
}

void DeinitPlayLines(void)
{
   if (PlayListFileName== NULL) return;
   StopBGPlay(0);

   while (--NumPlayLines >= 0)
       free(PlayLines[NumPlayLines]);

   free(PlayListFileName);
}

#define TEXT_TERM_WIDTH 80
extern char  *HelpComHeader;
extern char  *HelpComLines[];

void ShowComHelp(void)
{
    int l = strlen(HelpComHeader);
    char **arrPtr, *arrLine;

    puts(CLEAR_SCREEN);

    l = (TEXT_TERM_WIDTH-l) >> 1;
    while (--l >= 0) putchar(' ');
    printf("%s\n\n", HelpComHeader);


    arrPtr = HelpComLines;

    while ((arrLine = *arrPtr++) != NULL)
          printf ("%s\n", arrLine);
}

#ifndef X11
int SetColour(int  entry, char *ColourName)
{   int r, g, b;

#ifdef EGA
    if (VideoChipset == EGA) return 4;
#endif

    if (isdigit(*ColourName))
    {   long colourRef;

        sscanf(ColourName, "%lx", &colourRef);
        b = (int) (colourRef & 0xff);
        colourRef >>= 8;
        g = (int) (colourRef & 0xff);
        colourRef >>= 8;
        if (colourRef >= 0x100) return 4;
        r = (int) colourRef;
     }
     else
     {  FILE *f;
        const char *FileName;	
        char *s, cname[21];
        char line[81];

	FileName = COLOUR_NAME_FILE;
        f = fopen(FileName, "r");
#ifdef COLOUR_NAME_FILE_ALT
        if (!f)
        {  FileName = COLOUR_NAME_FILE_ALT; 
           f = fopen(FileName, "r");
        }
#endif
        if (!f)
        { printf ("\nFile %s was not found\n",
                      FileName);
          return 4;
        }

        while ((s = fgets(line, 80, f)) != NULL)
        {   if (*line == '!') continue;

            sscanf(line, "%d %d %d %20s", &r, &g, &b, cname);

            if (strcasecmp(ColourName, cname) == 0) break;
        }

        fclose(f);
        if (!s)
        { printf ("\nColour %s was not found in %s\n",
                  ColourName, FileName);
          return 4;
        }
     }

     r >>= 3;  b>>=3; g >>= 3;

     vga_setpalette(entry, r, g, b);
     return 0;
}
#endif
